/* $Id: path.c,v 1.16 1998/03/17 22:55:09 scott Exp $ */

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include "path.h"
#include "util.h"
#include "ranlib.h"

void pathDataAllocate(pathData *pd){
  /* Allocates memory for arrays in pathData structure based on the value
   * of pd->underliers, pd->times and pd->paths.
   */

  allocate2dDouble( &(pd->covariance), pd->underliers, pd->underliers);
  allocate1dDouble( &(pd->growthRate), pd->underliers);
  allocate1dDouble( &(pd->dividendRate), pd->underliers);
  allocate1dDouble(  &(pd->cov1), pd->underliers*pd->underliers);
  allocate1dDouble(  &(pd->fgrowth), pd->underliers);
  allocate1dDouble(  &(pd->parm), (pd->underliers*(pd->underliers+3)/2 +1));
  allocate3dDouble( &(pd->price), pd->times,pd->meshPaths,pd->underliers);
}

void pathDataDeallocate(pathData *pd){
  /* Deallocates memory for arrays in pathData structure based on the value
   * of pd->underliers, pd->times and pd->paths.
   */

  deallocate2dDouble( &(pd->covariance), pd->underliers, pd->underliers);
  deallocate1dDouble( &(pd->growthRate), pd->underliers);
  deallocate1dDouble( &(pd->dividendRate), pd->underliers);
  deallocate1dDouble(  &(pd->cov1), pd->underliers*pd->underliers);
  deallocate1dDouble(  &(pd->fgrowth), pd->underliers);
  deallocate1dDouble(  &(pd->parm), (pd->underliers*(pd->underliers+3)/2 +1));
  deallocate3dDouble( &(pd->price), pd->times,pd->meshPaths,pd->underliers);
}

void pathParamsInit(pathData *pp){
  /* Uses pp->underliers, pp->covariance[i][j], pp->growthRate[i], and
   * pp-phrase to initialize the private data and the RANLIB library.
   */

  int i,j,k;

  pp->constant = pow(2*M_PI,pp->underliers);
  for(i=0; i<pp->underliers; i++){
    pp->constant *= pp->covariance[i][i];
  }
  pp->constant = 1.0/sqrt(pp->constant);

  /* Allocate growth, same dimension as prices */
  allocate1dDouble(&(pp->growth),pp->underliers);


  /* Copy the 2d covariance matrix into the 1d scratch matrix */
  k = 0;
  for(i=0;i<pp->underliers;i++){
    for(j=0;j<pp->underliers;j++){
      pp->cov1[k] = pp->covariance[i][j];
      k++;
    }
  }

  /* Copy double growthRate array into double fgrowth array */
  /* Subtractthe dividend rate from the growth rate */
  for(i=0;i<pp->underliers;i++){
    pp->fgrowth[i] = pp->growthRate[i] - pp->dividendRate[i];
  }

  /* Give the data to RANLIB and let it do what it wants in parm */
  setgmn(pp->fgrowth,pp->cov1,pp->underliers,pp->parm);

  /* Hash the phrase into RNG seeds */
  phrtsd(pp->phrase,&(pp->seed1),&(pp->seed2));

  /* Initialize the RNG with the seeds */
  setall(pp->seed1,pp->seed2);

  /* Now whenever we want numbers from the distribution described in pp
   * we call genmn(pp->parm,vector,pp->cov1) where vector is a double array
   * of dimension [0:pp->underliers-1].  The random numbers go in vector
   */
}


void pathDataFill(pathData *pd){
  /* Uses pd->params, pd->paths, and pd->times to generate paths which are
   * placed in pd->price[p][t][u].  Make sure all these are initialized and
   * allocated first.  Each path starts with the value in pd->price[0][p][u]
   * so set these also.
   */

  int time, path, underlier,basePath;
  double *randomNum, std;

  randomNum = (double*)calloc(pd->underliers,sizeof(double));
  std = sqrt(pd->clusterVariance);

  /* pd->params describes the random numbers we get by calling
   * genmn(pd->params.parm, randomNum, pd->params.cov1).  To get the value in
   * the next time slice, we grow the values in the previous time slice by the
   * random numbers.  Each underlier gets its own growth rate.
   */

  for(time=1; time<pd->times; time++){
    for(path=0;path<pd->meshPaths;path++){
      if(path % pd->clusterSize == 0){
        genmn(pd->parm, randomNum, pd->cov1);
        for(underlier = 0; underlier < pd->underliers; underlier++){
          /*pd->price[time][path][underlier] = (1+randomNum[underlier])*  */
          /*  pd->price[time-1][path][underlier];  */
          pd->price[time][path][underlier] = 
            exp(randomNum[underlier]-pd->covariance[underlier][underlier]/2)*
            pd->price[time-1][path][underlier];
        }
        basePath = path;
      }else{
        genmn(pd->parm, randomNum, pd->cov1);
        for(underlier = 0; underlier < pd->underliers; underlier++){
          pd->price[time][path][underlier] =
            exp(randomNum[underlier]-pd->covariance[underlier][underlier]/2)*
            pd->price[time-1][basePath][underlier];
        }
      }
    }
  }

  free(randomNum);
}


double pathTransitionProb(pathData *pp, int time, int initialPath,
                          int finalPath){
  /* Calculates the probability of a transition from initial underlier proces
   * initialPrice[] to final underlier prices finalPrice[] over one timeslice.
   * Uses the statistics contained in pp.
   *
   * This is written for the case of independently distributed
   * underlier growth rates.  It should be expanded later.
   */

  int i;
  double exponential;

  for(i=0; i<pp->underliers; i++){
    pp->growth[i] = log(pp->price[time+1][finalPath][i]/
                        pp->price[time][initialPath][i]) +
      pp->covariance[i][i]/2;
  }

  /* Calculate the exponential term */
  exponential = 0;
  for(i=0; i<pp->underliers; i++){
    exponential += (pp->growth[i] - pp->fgrowth[i])*(pp->growth[i] - 
                   pp->fgrowth[i])/pp->covariance[i][i];
  }
  exponential = exp(-0.5*exponential);

  /* Put them together and we have our result */
  return(pp->constant*exponential);
}

double transProb(pathData *pp, double *price1, double *price2){
  int i;
  double exponential;

  for(i=0; i<pp->underliers; i++){
    pp->growth[i] = log(price2[i]/price1[i]) +
      pp->covariance[i][i]/2;
  }   

  /* Calculate the exponential term */
  exponential = 0;
  for(i=0; i<pp->underliers; i++){
    exponential += (pp->growth[i] - pp->fgrowth[i])*(pp->growth[i] -
                    pp->fgrowth[i])/pp->covariance[i][i];
  }
  exponential = exp(-0.5*exponential);

  /* Put them together and we have our result */
  return(pp->constant*exponential);
}        

void pathDataOut(pathData *pd,double t){ 
  int time, path, underlier,basePath; 
  double *x,*y; 
  int n,l; 
  int ierr; 
  n = pd->underliers*pd->times*pd->meshPaths; 
  x = (double*)calloc(n,sizeof(double)); 
  y = (double*)calloc(n,sizeof(double)); 
  l = 0; 

  for(underlier = 0; underlier < pd->underliers; underlier++){ 
    for(path=0;path<pd->meshPaths;path++){ 
      for(time=0; time<pd->times; time++){ 
        x[l]= (double)time; 
        y[l] = pd->price[time][path][underlier]; 
        /* printf("%i %i %lf %lf \n",path,time,x[l],y[l]);  */
        l++; 
      } 
    } 
  } 
  ierr = java_ser("paths",t,x,y,l); 

  free(x); 
  free(y); 
} 

/* Generates a random path and its antithetic */
void pathPair(pathData *pd, double **path, double **antiPath){
  int time, underlier;
  double  delta, antiGrowth;
  double *randomNum;

  allocate1dDouble(&randomNum,  pd->underliers);  

  for(underlier=0; underlier<pd->underliers; underlier++){
    path[0][underlier] = pd->price[0][0][underlier];
    antiPath[0][underlier] = pd->price[0][0][underlier];
  }
  
  for(time=1; time < pd->times; time++){
    genmn(pd->parm, randomNum, pd->cov1);
    for(underlier=0; underlier<pd->underliers; underlier++){
      path[time][underlier] =
        exp(randomNum[underlier]-pd->covariance[underlier][underlier]/2)*
        path[time-1][underlier];

      delta = randomNum[underlier] - pd->fgrowth[underlier]; 
      antiGrowth = pd->fgrowth[underlier] - delta; 

      antiPath[time][underlier] =
        exp(antiGrowth-pd->covariance[underlier][underlier]/2)*
        antiPath[time-1][underlier];
    }
  }
  deallocate1dDouble(&randomNum,  pd->underliers); 
}
